/*
 *
 *	wwwblock.c
 *
 *	H. Morrow Long
 *
 *	$Log: wwwblock.c,v $
 * Revision 1.2  1996/02/18  06:43:29  long
 * Added capability for 'implicit deny' policy -- if a file of allowable
 * URL prefixes exists, only those URL prefixes explicitly listed in the
 * file are allowed.  URLs with prefixes in the "censored-urls" file are
 * still denied.
 *
 * Revision 1.1  1996/02/18  04:04:36  long
 * Initial revision
 *
 * 
 *
 *	Solaris 2.*:	cc -o wwwblock wwwblock.c -lsocket -lnsl
 *	SunOS   4.*:	cc -o wwwblock wwwblock.c
 *
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <unistd.h>

#ifndef	ALLOWED_URLS_FILE
#define ALLOWED_URLS_FILE       "/usr/local/etc/allowed-urls"
#endif

#ifndef	CENSORED_URLS_FILE
#define	CENSORED_URLS_FILE	"/usr/local/etc/censored-urls"
#endif

#ifndef	WWW_PROXY_SERVER
	/* if you run a caching-proxy Web server on the same host */
#define	WWW_PROXY_SERVER	"localhost"
	/* otherwise, point to a caching-proxy Web server in your DMZ or 
		on the Internet */
/*	#define	WWW_PROXY_SERVER	"http-proxy.sub.dom.com"	*/
#endif

#ifndef	WWW_PROXY_PORT
#define	WWW_PROXY_PORT		8080
#endif

#define	RETURN_ERR	-1
#define	RETURN_FAIL	0
#define	RETURN_SUCCESS	1

#define	MAX_FDS		32

int             Debug;
int             Verbose;

main(ArgC, ArgV)
	int             ArgC;
	char          **ArgV;
{
	int             Index;
	int             SubIndex;

	for (Index = 1; (Index < ArgC) && (ArgV[Index][0] == '-'); Index++)
	    for (SubIndex = 1; ArgV[Index][SubIndex]; SubIndex++)
		switch (ArgV[Index][SubIndex])
		{
		case 'd':
			Debug++;
			break;
		case 'v':
			Verbose++;
			break;
		default:
			(void) fprintf(stderr,
		"Usage:  wwwblock\n");
			exit(1);
		}

		openlog("wwwblock", LOG_PID, LOG_DAEMON);
		(void) WWW_Proxy_Setup();
		closelog();
	exit(0);
}

WWW_Proxy_Setup()
{
	unsigned        Port;
	char           *Host;
	struct hostent *HostEntryPointer;
	struct sockaddr_in SocketInetAddr;
	struct hostent  TargetHost;
	struct in_addr  TargetHostAddr;
	char           *AddressList[1];
	char            NameBuffer[128];

	extern int      inet_addr();
	extern char    *rindex();

	Host = WWW_PROXY_SERVER;
	if (Host == NULL)
		return (RETURN_FAIL);
	HostEntryPointer = gethostbyname(Host);
	if (HostEntryPointer == NULL)
		{
		TargetHostAddr.s_addr = inet_addr(Host);
		if (TargetHostAddr.s_addr == -1)
			{
			(void) fprintf(stderr, "unknown host: %s\n", Host);
			return (RETURN_FAIL);
			}
		(void) strcpy(NameBuffer, Host);
		TargetHost.h_name = NameBuffer;
		TargetHost.h_addr_list = AddressList, TargetHost.h_addr = 
			(char *) &TargetHostAddr;
		TargetHost.h_length = sizeof(struct in_addr);
		TargetHost.h_addrtype = AF_INET;
		TargetHost.h_aliases = 0;
		HostEntryPointer = &TargetHost;
		}
	SocketInetAddr.sin_family = HostEntryPointer->h_addrtype;
	bcopy(HostEntryPointer->h_addr, (char *) &SocketInetAddr.sin_addr,
		HostEntryPointer->h_length);


	(void) WWW_Proxy(WWW_PROXY_PORT, HostEntryPointer, SocketInetAddr);
	return (RETURN_SUCCESS);
}

WWW_Proxy(Port, HostEntryPointer, SocketInetAddr)
	unsigned        Port;
	struct hostent *HostEntryPointer;
	struct sockaddr_in SocketInetAddr;
{
	fd_set		FDs;
	char		Byte;
	int		N;
	int		R;
	FILE *WWW_Proxy_Incoming;
	FILE *WWW_Proxy_Outgoing;
	char            Buffer[BUFSIZ];
	int             SocketDescriptor;
	struct servent *ServiceEntryPointer;


	SocketInetAddr.sin_port = htons(Port);
	SocketDescriptor = socket(AF_INET, SOCK_STREAM, 6);
	if (SocketDescriptor < 0)
		{
		perror("socket");
		return (RETURN_ERR);
		}
	if (Debug)
		{
		(void) printf("Host %s, Port %d ", HostEntryPointer->h_name,
			      Port);
		if ((ServiceEntryPointer = getservbyport(htons(Port), "tcp")) !=
		    (struct servent *) NULL)
			(void) printf(" (\"%s\" service) ",
				      ServiceEntryPointer->s_name);
		(void) printf("connection ... ");
		(void) fflush(stdout);
		}
	if (connect(SocketDescriptor, (char *) &SocketInetAddr,
		    sizeof(SocketInetAddr)) < 0)
		{
		if (Debug)
			(void) printf("NOT open.\n");
		if (Debug)
			syslog(LOG_INFO|LOG_LOCAL2, "error: connect(): %m");
		}
	else
		{
		if (Debug)
			{
			(void) printf("Host %s, Port %d ",
				      HostEntryPointer->h_name, Port);
			if ((ServiceEntryPointer = getservbyport(htons(Port),"tcp")) !=
			    (struct servent *) NULL)
				(void) printf(" (\"%s\" service) ",
					      ServiceEntryPointer->s_name);
			(void) printf("connection ... ");
			(void) printf("open.\n");
			(void) fflush(stdout);
			}

		N = 0;
		WWW_Proxy_Outgoing = fdopen(SocketDescriptor, "w");
	        while ( (read(STDIN_FILENO, &Byte, 1) == 1) &&
			(  Byte != '\012' ) && 
			(  Byte != '\015' ) )
			{
				Buffer[N++] = Byte;
				putc ( Byte, WWW_Proxy_Outgoing );
			if (Debug)
				{
				 printf("inside read loop\n");
				(void) fflush(stdout);
				}
			}
	
		Buffer[N] = '\0';
		if (Check_URL(Buffer) == RETURN_FAIL ) 
			{
			printf("HTTP/1.0 403 Forbidden - Blocked site: %s\r\n",
				Buffer);
			printf("Server: WWWBlock 1.0\r\n");
			printf("MIME-version: 1.0\r\n");
			printf("Content-type: text/html\r\n");
			printf("\r\n");
			printf("<TITLE>WWWBlocked site: %s</TITLE>\r\n",Buffer);
			printf("<H1>Access Forbidden!</H1>\r\n");
			printf("<H2>Blocked site: %s</H2>\r\n" , Buffer);
			printf("\r\n");
			(void) fflush(stdout);
			exit(255);
			}

		putc ( '\015', WWW_Proxy_Outgoing );
		putc ( '\012', WWW_Proxy_Outgoing );
		(void) fflush(WWW_Proxy_Outgoing);

		if (Debug)
			{
			(void) printf("now reading from remote host.\n");
			(void) fflush(stdout);
			}

		for (;;)
			{
			FD_ZERO(&FDs);
			FD_SET(SocketDescriptor, &FDs);
			FD_SET(STDIN_FILENO, &FDs);

			if (!(R = select(MAX_FDS, &FDs,
				(fd_set *)STDIN_FILENO,(fd_set *)0,0) ))
				break;
			else if (R < 0 ) continue;

			if ( FD_ISSET( SocketDescriptor, &FDs ) )
				{
				if ( (R = recv(SocketDescriptor, Buffer, 
					sizeof(Buffer), 0)) < 1) break;
				write(STDOUT_FILENO, Buffer, R);
				}
			if ( FD_ISSET ( STDIN_FILENO, &FDs ) )
				{
				if ( (R = read(STDIN_FILENO, Buffer, 
					sizeof(Buffer))) < 1) break;
				send(SocketDescriptor, Buffer, R, 0);
				}
			}


		(void) fclose ( WWW_Proxy_Outgoing );
		}

	(void) close(SocketDescriptor);
	return (RETURN_SUCCESS);
}

Check_URL(Buffer)
char	*Buffer;
{
FILE	*AllowedURLs;
/* ALLOWED_URLS_FILE */
FILE	*CensoredURLs;
char	*oldURL;
char	*newURL;
char	Line[4096];
int	i;

	/*
	 *	Parse HTTP request, extract URL, downcase it for matching.
	 */

	oldURL = Buffer;
	while ( ( *oldURL ) && ( ! ( isspace(*oldURL)) ) ) oldURL++;
	if ( ( *oldURL ) && ( isspace(*oldURL)) ) oldURL++;
	if ( ! ( newURL = malloc(strlen(oldURL + 1)) ) )
		{
		perror( "malloc: strlen(oldURL + 1)" );
		syslog(LOG_INFO|LOG_LOCAL2, "error: malloc(oldURL + 1): %m");
		
		exit(255);
		}
	for ( i = 0; oldURL[i] && ( ! ( isspace( oldURL[i] ) )); i++)
		newURL[i] = isupper( oldURL[i] ) ? tolower ( oldURL[i] ) :
			 oldURL[i];
	newURL[i] = '\0';

	/*
	 *	Check for prefix matching against allowed URLs in file.
	 *	If an allowed URLs file exists then the current HTTP 
	 *	request URL must match one of the prefixes in the file.
	 */	


	if (( AllowedURLs = fopen( ALLOWED_URLS_FILE, "r" )) != (FILE *)0)
	    {
	    while ( fgets ( Line, sizeof(Line), AllowedURLs )
		 != (char *)0 )
		{
		for ( i = 0; Line[i] && (Line[i] != '\n'); i++)
			Line[i] = isupper( Line[i] ) ? tolower ( Line[i] ) :
				 Line[i];
		Line[i] = '\0';
		if ( ! (strncmp( newURL, Line, strlen(Line))) )
			break;
		}
	    if ( feof( AllowedURLs ) )
			return(RETURN_FAIL ) ;
	    fclose( AllowedURLs );
	    }
		

	/*
	 *	Check for prefix matching against censored URLs in file.
	 */	

	if (( CensoredURLs = fopen( CENSORED_URLS_FILE, "r" )) == (FILE *)0)
		{
		perror( CENSORED_URLS_FILE );
		syslog(LOG_INFO|LOG_LOCAL2, "error: open(CENSORED_URLS_FILE): %m");
		
		exit(255);
		}
	while ( fgets ( Line, sizeof(Line), CensoredURLs ) != (char *)0 )
		{
		for ( i = 0; Line[i] && (Line[i] != '\n'); i++)
			Line[i] = isupper( Line[i] ) ? tolower ( Line[i] ) :
				 Line[i];
		Line[i] = '\0';
		if ( ! (strncmp( newURL, Line, strlen(Line))) )
			return(RETURN_FAIL ) ;
		}
		

	fclose( CensoredURLs );
	return(RETURN_SUCCESS ) ;
}

